#ifndef MAPREDUCE_H
#define MAPREDUCE_H

/**
 * @file mapreduce.h
 * @brief Header defining mapreduce library interface.
 */

#ifdef __cplusplus
#  include <cstdlib>
#else
#  include <stdlib.h>
#  include <stdbool.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @struct mapreduce_input_reader
 * @brief Type of object for reading from maping task's input file
 */
struct mapreduce_input_reader;

/**
 * @struct mapreduce_intermediate_reader
 * @brief Type of object for reading from intermediate file by reducing task
 */
struct mapreduce_intermediate_reader;

/**
 * @struct mapreduce_output_writer
 * @brief Type of object for writing to reducing task's input file
 */
struct mapreduce_output_writer;

/**
 * @struct mapreduce_intermediate_writer
 * @brief Type of object for writing to intermediate file by mapping task
 */
struct mapreduce_intermediate_writer;

/**
 * @struct mapreduce_specification
 * @brief Details of map-reduce algorithm that have to be
 *        specified by user.
 */
struct mapreduce_specification {
        bool (*map)(struct mapreduce_input_reader* in_r,
                    struct mapreduce_intermediate_writer* im_w);
        bool (*reduce)(struct mapreduce_intermediate_reader* im_r,
                       struct mapreduce_output_writer* out_w);
        const char** inputs;
        const char** outputs;
        size_t inputs_num;
        size_t outputs_num;
};

/**
 * @enum mapreduce_error_code
 * @brief Error code of mapreduce function in
 *        case of its unsuccesful termination.
 */
enum mapreduce_error_code {
        MAPREDUCE_ERROR_NONE,
        MAPREDUCE_ERROR_MAP_FAILURE,
        MAPREDUCE_ERROR_REDUCE_FAILURE,
        MAPREDUCE_ERROR_MAP_TERMINATION,
        MAPREDUCE_ERROR_REDUCE_TERMINATION,
        MAPREDUCE_ERROR_INTERNAL
};

/**
 * @struct mapreduce_error
 * @brief Specifies reason of map-reduce's
 *        unsuccesful termination.
 */
struct mapreduce_error {
        enum mapreduce_error_code code;
};

/**
 * @brief Begins execution of map-reduce algorithm.
 * @param spec Describes usage-specific details of algorithm
 *             execution.
 * @param err If false is returned, describes error that occured.
 *            Otherwise, its value on return is unspecified.
 * @return true if mapreduce was successfuly executed,
 *         false otherwise.
 */
bool mapreduce(const struct mapreduce_specification* spec,
               struct mapreduce_error* err);

/**
 * @brief Reads from map-reduce input file.
 * @param in_r Reader used for reading operation.
 * @param data Buffer for data read from input.
 * @param size Size of data to read. Changed to size of valid data in
 *             given buffer on return. Becomes 0 on EOF. Its value
 *             after unsuccessful reading is unspecified.
 * @return true if reading operation was successful, false otherwise.
 */
bool mapreduce_read_input(struct mapreduce_input_reader* in_r,
                          void* data, size_t* size);

/**
 * @brief Reads from map-reduce intermediate file.
 * @param im_r Reader used for reading operation.
 * @param data Buffer for data read from input.
 * @param size Size of data to read. Changed to size of valid data in
 *             given buffer on return. Becomes 0 on EOF. Its value
 *             after unsuccessful reading is unspecified.
 * @return true if reading operation was successful, false otherwise.
 */
bool mapreduce_read_intermediate(struct mapreduce_intermediate_reader* im_r,
                                 void* data, size_t* size);

/**
 * @brief Obtains number of partitions to prepare by map task.
 * @param im_w Writer object.
 * @return Number of partitions to which data can be written.
 */
size_t mapreduce_parts_num(const struct mapreduce_intermediate_writer* im_w);

/**
 * @brief Write to map-reduce intermediate file.
 * @param im_w Writer used for writing operation.
 * @param data Buffer with data written to partition.
 * @param size Size of data to write. Changed to size of data written
 *             to file on return. Its value
 *             after unsuccessful writing is unspecified.
 * @param part_num Number of partition of intermediate file to write to.
 * @return true if writing operation was successful, false otherwise.
 */
bool mapreduce_write_part(struct mapreduce_intermediate_writer* im_w,
                          const void* data, size_t* size, size_t part_num);
/**
 * @brief Write to map-reduce output file.
 * @param out_w Writer used for writing operation.
 * @param data Buffer with data written to output.
 * @param size Size of data to write. Changed to size of data written
 *             to file on return. Its value
 *             after unsuccessful writing is unspecified.
 * @return true if writing operation was successful, false otherwise.
 */
bool mapreduce_write_out(struct mapreduce_output_writer* out_w,
                         const void* data, size_t* size);

#ifdef __cplusplus
}
#endif

#endif // MAPREDUCE_H

