/**
 * @file master.c
 * @brief Implements master task's functionality.
 */

#include <iso646.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <sys/wait.h>
#include <sys/sem.h>

#include <mapreduce/master.h>
#include <mapreduce/mapper.h>
#include <mapreduce/reducer.h>
#include <mapreduce/partitions.h>
#include <mapreduce/task.h>
#include <mapreduce/logger.h>

/**
 * @brief Temporary file name templates.
 */
static const char* const part_tmpl = "/tmp/mapreduce-part-XXXXXX";
static const char* const im_tmpl = "/tmp/mapreduce-intermediate-XXXXXX";
static const char* const tasks_tmpl = "/tmp/mapreduce-tasks-XXXXXX";
static const size_t part_tmpl_size = (sizeof "/tmp/mapreduce-part-XXXXXX") + sizeof '\0';
static const size_t im_tmpl_size = (sizeof "/tmp/mapreduce-intermediate-XXXXXX") + sizeof '\0';
static const size_t tasks_tmpl_size = (sizeof "/tmp/mapreduce-tasks-XXXXXX") + sizeof '\0';

/**
 * @brief Helper functions declarations.
 */

// Checks the intermediate file partitioning decription produced by
// mapper task with given pid and proceses it to reduce tasks.
static bool notify_intermediate_ready(const struct master_data* master,
				      pid_t child, struct mapreduce_error* err);

// Initializes Master process data structures and its subprocesses.
static struct master_data master_init( \
	const struct mapreduce_specification* spec,
	struct mapreduce_error* err);

// Deinitializes Master process data structures.
static void master_destroy(struct master_data* master);

// Initializes mapper descriptor and launches mapping process.
static struct mapper_descriptor mapper_init( \
	const struct mapreduce_specification* spec,
	const char* in_filename,
	struct mapreduce_error* err);

// Deinitializes mapper descriptor: closes its file descriptors etc.
static void mapper_destroy(struct mapper_descriptor* mapper);

// Initializes mapper descriptor and launches mapping process.
static struct reducer_descriptor reducer_init( \
	const struct mapreduce_specification* spec,
	const char* im_filenames[],
	const char* out_filename,
	struct mapreduce_error* err);

// Deinitializes mapper descriptor: closes its file descriptors etc.
static void reducer_destroy(struct reducer_descriptor* reducer);

// Returns true if every worker task did complete its work.
static bool are_all_tasks_completed(const struct master_data* master);

// Returns true if worker process with given pid is mapping process.
static bool is_mapper(const struct master_data* master, pid_t worker_pid);

// Returns descriptor of mapping process by its pid.
static const struct mapper_descriptor* get_mapper( \
	const struct master_data* master, pid_t worker_pid);

// Returns ordinal number of intermediate file with output produced by
// mapping process with given worker_id.
static size_t get_im_id(const struct master_data* master, pid_t worker_pid);

// Tells Master process that worker process with given pid did complete its work.
static void mark_worker_done(struct master_data* master, pid_t worker_pid);

// Reads intermediate file partitioning description to given structure.
static bool read_partitions(const struct mapper_descriptor* mapper,
			    struct partitions* parts);

// Distributes tasks among reducing worker processes basing on given
// intermediate file partitioning description.
static bool write_tasks(const struct master_data* master,
			const size_t intermediate_id,
			const struct partitions* parts);

// Writes task to reducer processes "tasks" file.
static bool write_task_line(const struct reducer_descriptor* reducer,
			    const size_t intermediate_id,
			    off_t offset, size_t size);


bool
master_task(const struct mapreduce_specification* const spec,
	    struct mapreduce_error* const err)
{
	logger_init();
	logger_event("Master initialized\n");
	struct master_data master = master_init(spec,err);
	bool successful = true;
	while (not are_all_tasks_completed(&master)) {
		int status = 0;
		const pid_t child = wait(&status);
		const bool mapping_completed = is_mapper(&master, child);
		if (WIFEXITED(status)) {
			const int exit_status = WEXITSTATUS(status);
			const bool exit_success = exit_status == 0;
			if (not exit_success) {
				err->code = exit_status;
				successful = false;
				break;
			}
		} else if (WIFSTOPPED(status)) {
			continue;
		} else { // WIFSIGNALED(status)
			err->code = mapping_completed ?			\
				MAPREDUCE_ERROR_MAP_TERMINATION :	\
				MAPREDUCE_ERROR_REDUCE_TERMINATION;
			successful = false;
			break;
		}
		mark_worker_done(&master, child);
		if (mapping_completed) {
			const bool done = notify_intermediate_ready(&master,
								    child, err);
			if (not done) {
				successful = false;
				break;
			}
		}
	}
	master_destroy(&master);
	return successful;
}

static bool
notify_intermediate_ready(const struct master_data* const master,
			  const pid_t child, struct mapreduce_error* const err)
{
	logger_event("Notify reducer task\n");
	const struct mapper_descriptor* const mapper = \
		get_mapper(master, child);
	struct partitions parts = partitions_init(master->reducers_num);
	const bool read = read_partitions(mapper, &parts);
	if (not read) {
		err->code = MAPREDUCE_ERROR_INTERNAL;
		partitions_destroy(&parts);
		return false;
	}
	const size_t im_id = get_im_id(master, child);
	const bool written = write_tasks(master, im_id,
					 &parts);
	if (not written) {
		err->code = MAPREDUCE_ERROR_INTERNAL;
		partitions_destroy(&parts);
		return false;
	}
	partitions_destroy(&parts);
	return true;
}

static struct master_data
master_init(const struct mapreduce_specification* const spec,struct mapreduce_error* err)
{
	logger_event("Master task init\n");
	const size_t map_num = spec->inputs_num;
	const size_t reduce_num = spec->outputs_num;
	const struct master_data init = {
		.mappers = calloc(map_num, sizeof init.mappers[0]),
		.reducers = calloc(reduce_num, sizeof init.reducers[0]),
		.mappers_num = map_num,
		.reducers_num = reduce_num
	};
	const char** const im = calloc(map_num, sizeof im[0]);
	for (size_t i = 0; i < map_num; i++) {
		init.mappers[i] = mapper_init(spec, spec->inputs[i],err);
		im[i] = init.mappers[i].intermediate_filename;
	}
	for (size_t i = 0; i < reduce_num; i++)
		init.reducers[i] = reducer_init(spec, im, spec->outputs[i],err);
	free(im);
	return init;
}

static void
master_destroy(struct master_data* const master)
{
	logger_event("Master task destroy\n");
	logger_destroy();
	for (size_t i = 0; i < master->mappers_num; i++)
		mapper_destroy(&master->mappers[i]);
	for (size_t i = 0; i < master->reducers_num; i++)
		reducer_destroy(&master->reducers[i]);
	free(master->mappers);
	free(master->reducers);
	master->mappers = NULL;
	master->reducers = NULL;
	master->mappers_num = 0;
	master->reducers_num = 0;
}

static struct mapper_descriptor
mapper_init(const struct mapreduce_specification* const spec,
	    const char* const in_filename,struct mapreduce_error* err)
{
	logger_event("Mapper task init\n");
	char part[part_tmpl_size];
	char im[im_tmpl_size];
	const int part_fd = mkstemp(strcpy(part, part_tmpl));
	const int im_fd = mkstemp(strcpy(im, im_tmpl));
	const pid_t pid = fork();
	if (pid == 0) {
		const char* const in = in_filename;
		const size_t partitions_num = spec->outputs_num;
		const int in_fd = open(in, O_RDONLY);
		const struct mapper_env env = {
			.in_fd = in_fd,
			.im_fd = im_fd,
			.part_fd = part_fd,
			.partitions_num = partitions_num,
			.map = spec->map	
		};
		mapper_task(&env,err);
		close(im_fd);
		close(part_fd);
		close(in_fd);
		exit(err->code);
	}
	char* const p_alloc = calloc(part_tmpl_size, sizeof part[0]);
	char* const i_alloc = calloc(im_tmpl_size, sizeof im[0]);
	const struct mapper_descriptor init = {
		.pid = pid,
		.part_filename = strcpy(p_alloc, part),
		.intermediate_filename = strcpy(i_alloc, im),
		.completed = false
	};
	close(im_fd);
	close(part_fd);
	return init;
}

static void
mapper_destroy(struct mapper_descriptor* const mapper)
{
	logger_event("Mapper destroy\n");
	unlink(mapper->part_filename);
	unlink(mapper->intermediate_filename);
	free(mapper->part_filename);
	free(mapper->intermediate_filename);
	mapper->pid = -1;
	mapper->part_filename = NULL;
	mapper->intermediate_filename = NULL;
	mapper->completed = true;
}

static struct reducer_descriptor
reducer_init(const struct mapreduce_specification* const spec,
	     const char* im_filenames[],
	     const char* const out_filename,struct mapreduce_error* err)
{
	logger_event("Reducer task init\n");
	char tasks[tasks_tmpl_size];
	const int tasks_sem = semget(IPC_PRIVATE, 1, 0600);
	const int tasks_fd_tmp = mkstemp(strcpy(tasks, tasks_tmpl));
	const pid_t pid = fork();
	const int tasks_fd = open(tasks, O_RDWR | O_APPEND);
	close(tasks_fd_tmp);
	if (pid == 0) {
		const size_t im_num = spec->inputs_num;
		const char* const o = out_filename;
		const int out_fd = open(o, O_CREAT | O_WRONLY | O_TRUNC |O_APPEND, S_IRUSR | S_IWUSR);
		const struct reducer_env env = {
			.out_fd = out_fd,
			.tasks_fd = tasks_fd,
			.tasks_sem = tasks_sem,
			.im_filenames = im_filenames,
			.im_num = im_num,
			.reduce = spec->reduce
		};
		reducer_task(&env,err);
		close(tasks_fd);
		close(out_fd);
		exit(err->code);
	}
	char* const t_alloc = calloc(tasks_tmpl_size, sizeof tasks[0]);
	const struct reducer_descriptor init = {
		.pid = pid,
		.tasks_filename = strcpy(t_alloc, tasks),
		.tasks_fd = tasks_fd,
		.tasks_sem = tasks_sem,
		.completed = false
	};
	return init;
}

static void
reducer_destroy(struct reducer_descriptor* const reducer)
{
	logger_event("Reducer task destroy\n");
	close(reducer->tasks_fd);
	unlink(reducer->tasks_filename);
	semctl(reducer->tasks_sem, 0, IPC_RMID);
	free(reducer->tasks_filename);
	reducer->pid = -1;
	reducer->tasks_filename = NULL;
	reducer->tasks_fd = -1;
	reducer->tasks_sem = -1;
	reducer->completed = true;
}

static bool
are_all_tasks_completed(const struct master_data* const master)
{
	for (size_t i = 0; i < master->mappers_num; i++)
		if (master->mappers[i].completed == false)
			return false;
	for (size_t i = 0; i < master->reducers_num; i++)
		if (master->reducers[i].completed == false)
			return false;
	return true;
}

static bool
is_mapper(const struct master_data* const master, const pid_t worker_pid)
{
	for (size_t i = 0; i < master->mappers_num; i++)
		if (master->mappers[i].pid == worker_pid)
			return true;
	return false;
}

static const struct mapper_descriptor*
get_mapper(const struct master_data* const master, const pid_t worker_pid)
{
	for (size_t i = 0; i < master->mappers_num; i++)
		if (master->mappers[i].pid == worker_pid)
			return &master->mappers[i];
	return NULL;
}

static size_t
get_im_id(const struct master_data* const master, const pid_t worker_pid)
{
	for (size_t i = 0; i < master->mappers_num; i++)
		if (master->mappers[i].pid == worker_pid)
			return i;
	return master->mappers_num;
}

static void
mark_worker_done(struct master_data* const master, const pid_t worker_pid)
{
	for (size_t i = 0; i < master->mappers_num; i++)
		if (master->mappers[i].pid == worker_pid)
			master->mappers[i].completed = true;
	for (size_t i = 0; i < master->reducers_num; i++)
		if (master->reducers[i].pid == worker_pid)
			master->reducers[i].completed = true;
}

static bool
read_partitions(const struct mapper_descriptor* const mapper,
		struct partitions* const parts)
{
	const int fd = open(mapper->part_filename, O_RDONLY);
	bool no_problems = true;
	for (size_t i = 0; i < parts->parts_num; i++) {
		const ssize_t result = read(fd, &parts->parts[i],
					    sizeof parts->parts[0]);
		no_problems = (result == -1) ? false : no_problems;
	}
	return no_problems;
}

static bool
write_tasks(const struct master_data* const master,
	    const size_t intermediate_id,
	    const struct partitions* const parts)
{
	bool no_problems = true;
	for (size_t i = 0; i < parts->parts_num; i++) {
		const bool result = write_task_line(&master->reducers[i],
						    intermediate_id,
						    parts->parts[i].offset,
						    parts->parts[i].size);
		no_problems = result ? no_problems : false;
	}
	return no_problems;
}

static bool
write_task_line(const struct reducer_descriptor* const reducer,
		const size_t intermediate_id,
		const off_t offset, const size_t size)
{
	struct sembuf op = {
		.sem_num = 0,
		.sem_op = +1,
		.sem_flg = SEM_UNDO
	};
	const struct task task = {
		.intermediate_id = intermediate_id,
		.part_start = offset,
		.part_size = size
	};
	const ssize_t result = write(reducer->tasks_fd, &task, sizeof task);
	semop(reducer->tasks_sem, &op, 1);
	return result not_eq -1;
}
