/**
 * @file reducer.c
 * @brief Implements reducer task's functionality.
 */

#include <stdlib.h>
#include <unistd.h>
#include <iso646.h>

#include <sys/wait.h>
#include <sys/sem.h>

#include <mapreduce/reducer.h>
#include <mapreduce/logger.h>

/**
 * @brief Helper functions declarations.
 */

// Receive Reducer Task by reading Reducer Task file.
static bool receive_task(const struct reducer_env* const environment, struct reducer_data* data);
//Initialize Reducer data structure by using reducer_env strucutre from Master process
static bool prepare_reducer(const struct reducer_env* const environment, struct reducer_data* data);


static bool
receive_task(const struct reducer_env* const env, struct reducer_data* const data)
{
	struct sembuf op = {
		.sem_num = 0,
		.sem_op = -1,
		.sem_flg = SEM_UNDO
	};
	semop(env->tasks_sem, &op, 1);
	logger_event("Reducer received task!\n");
	const ssize_t data_read = read(env->tasks_fd, &data->reduce_task, sizeof data->reduce_task);
	return data_read == sizeof data->reduce_task;
}

static bool
prepare_reducer(const struct reducer_env* const env, struct reducer_data* const data)
{
	const int fd = open(env->im_filenames[data->reduce_task.intermediate_id], O_RDONLY);
	const off_t offset = lseek(fd, data->reduce_task.part_start, SEEK_SET);	
	if (offset not_eq data->reduce_task.part_start) {
		close(fd);
		return false;
	}
	data->im_r.fd = fd;
	data->im_r.partition_size = data->reduce_task.part_size;
	data->out_w.fd = env->out_fd;
	return true;
}

void
reducer_task(const struct reducer_env* const env, struct mapreduce_error* err)
{
	logger_event("Reducer task executed\n");
	struct reducer_data data;
	for (size_t i = 0; i < env->im_num; ++i) {
		if (not receive_task(env,&data)) {
			logger_event("Reducer receive task error occurred!\n");
			err->code = MAPREDUCE_ERROR_INTERNAL;	
			return;
		}
		if (not prepare_reducer(env,&data)) {
			logger_event("Reducer prepare error occurred!\n");
			err->code = MAPREDUCE_ERROR_INTERNAL;
			return;
		}	
		if (not env->reduce(&data.im_r,&data.out_w)) {
			logger_event("Reducer reduce() error occurred!\n");
			err->code = MAPREDUCE_ERROR_REDUCE_FAILURE;
			return;
		}
		logger_event("Reducer one task success\n");
	}
	logger_event("Reducers all tasks success\n");
}

