/**
 * @file ImageTest.cpp
 * @brief File contains definition of ImageTest test class.
 */

#include <mapreduce/tests/ImageTest.hpp>

using namespace std;

namespace MapReduceImageTest
{
	/**
	 * MapReduce user defined map function.
	 * @param	in_r is pointer to input_reader struct used to read mapper file.
	 * @param	im_w is pointer to intermediate_writer struct used to write mapping parts.
	 * @return	true if mapping completed without any errors.
	 */
	bool
	map(struct mapreduce_input_reader* const in_r, struct mapreduce_intermediate_writer* const im_w)
	{		
		char inputBuffer[32];
		size_t inputBufferSize = (sizeof inputBuffer)/(sizeof inputBuffer[0]);

		const bool readSuccess = mapreduce_read_input(in_r,&inputBuffer,&inputBufferSize); 

		if(not readSuccess)
			return false;


		char secondPart[16];
		char firstPart[16];

		for(int i = 0; i < 16; ++i)
		{
			firstPart[i] = '0';
			secondPart[i] = '0';
		}

		size_t firstSize = 16;
		size_t secondSize = 16;

		memcpy(firstPart,inputBuffer,16);
		memcpy(secondPart,inputBuffer,16);
		bool writeSuccess = mapreduce_write_part(im_w, &firstPart, &firstSize, 0);
	
		if(not writeSuccess)
			return false;
	
		writeSuccess = mapreduce_write_part(im_w,&secondPart, &secondSize, 1);

		if(not writeSuccess)
			return false;
	
  		return true;
	}

	/**
	 * MapReduce user defined reduce function.
	 * @param	im_r is pointer to intermediate_reader struct used to read mappend files.
	 * @param	out_w is pointer to output_writer struct used to write reduce result to output file.
	 * @return	true if mapping completed without any errors.
	 */
	bool
	reduce(struct mapreduce_intermediate_reader* const im_r, struct mapreduce_output_writer* const out_w)
	{
		char buffer[16];

		for(int i = 0; i < 16; ++i)
			buffer[i] = '0';
	
		size_t bufferSize = (sizeof buffer)/(sizeof buffer[0]);

		const bool readSuccess = mapreduce_read_intermediate(im_r,&buffer,&bufferSize); 


		if(not readSuccess)
			return false;

		if(buffer[0] != '0')
			buffer[15] = '\0';

		for(int i = 0; i < 16; ++i)
			if(buffer[i] == '2')
			{
				char error[] = {'E','r','r','o','r'};
				size_t size = (sizeof error)/(sizeof error[0]);
				int write_success = mapreduce_write_out(out_w,&error, &size);
	
				if(not write_success)
					return false;
			}

		return true;
	}

	/**
 	 * Class default constructor, allocate memory and prepare test.
	 */
	ImageTest::ImageTest()
	{
		srand(time(NULL));
		int bug = rand() % 10 + 1;
		int fileID = rand() % 20 + 1;

		FILE* file;

		inputs[0] = "0.raw";
		inputs[1] = "1.raw";
		inputs[2] = "2.raw";
		inputs[3] = "3.raw";
		inputs[4] = "4.raw";
		inputs[5] = "5.raw";
		inputs[6] = "6.raw";
		inputs[7] = "7.raw";
		inputs[8] = "8.raw";
		inputs[9] = "9.raw";

		inputs[10] = "10.raw";
		inputs[11] = "11.raw";
		inputs[12] = "12.raw";
		inputs[13] = "13.raw";
		inputs[14] = "14.raw";
		inputs[15] = "15.raw";
		inputs[16] = "16.raw";
		inputs[17] = "17.raw";
		inputs[18] = "18.raw";
		inputs[19] = "19.raw";

		outputs[0] = "100.raw";
		outputs[1] = "101.raw";
		outputs[2] = "102.raw";
		outputs[3] = "103.raw";
		outputs[4] = "104.raw";
		outputs[5] = "105.raw";
		outputs[6] = "106.raw";
		outputs[7] = "107.raw";
		outputs[8] = "108.raw";
		outputs[9] = "109.raw";

		outputs[10] = "110.raw";
		outputs[11] = "111.raw";
		outputs[12] = "112.raw";
		outputs[13] = "113.raw";
		outputs[14] = "114.raw";
		outputs[15] = "115.raw";
		outputs[16] = "116.raw";
		outputs[17] = "117.raw";
		outputs[18] = "118.raw";
		outputs[19] = "119.raw";

		properResult = true;
		for(int i = 0; i < 20; ++i)
		{
			file = fopen(inputs[i], "wb");


			if(i == fileID && bug  >= 4)
			{
				fwrite("111111110002000011111111000000\n",1,sizeof("111111110002000011111111000000\n"),file);
				properResult = false;
			}
			else
				fwrite("111111110000000011111111000000\n",1,sizeof("111111110000000011111111000000\n"),file);
			
			fclose(file);
		}		

		for(int i = 0; i < 20; ++i)
		{	
			FILE* file = fopen(outputs[i], "wb");
			fclose(file);
		}
		prepareTest();
	
	}

	/**
	 * Class default destructor, free allocated memory.
	 */
	ImageTest::~ImageTest()
	{

	}
	
	/**
	 * Method is used to prepare MapReduce specification of test case.
	 */
	void
	ImageTest::prepareTest()
	{
		const size_t inputsNum = (sizeof inputs)/(sizeof inputs[0]);
		const size_t outputsNum = (sizeof outputs)/(sizeof outputs[0]);

		spec.map = &map;
		spec.reduce = &reduce;
  		spec.inputs = inputs;
  		spec.outputs = outputs;
  		spec.inputs_num = inputsNum;
  		spec.outputs_num = outputsNum;
	}

	/**
	 * Method is used to run Image bitmap validation test.
	 */
	void
	ImageTest::runTest()
	{
		const bool ok = mapreduce(&spec, &error);
	
		bool result = properResult;

		for(int i = 0; i < 20; ++i)
		{
			ifstream file;
			string error;

			file.open(outputs[i]);
			file >> error;

			if(!strcmp(error.c_str(), "Error"))
				result = false;
	
			file.close();
		}

		cout << "Bitmap was " << (properResult ? "ok " : "not ok") << endl;

		cout << "MapReduce result was " <<(result ? "ok" : "not ok") << endl;

		if (not ok) {
			cout << "ImageTest ok: map-reduce failed. Code: " << error.code << endl;
		}
		
		if(properResult == result)
			cout << "ImageTest ok: Bitmap and MapReduce results are matching"
			     << endl;
		else
			cout << "ImageTest not okk: Bitmap and MapReduce results are not matching"
			     << endl;
	}	
}
