/**
 * @file PatternSearch.cpp
 * @brief File contains definition of PatternSearch test class.
 */

#include <mapreduce/tests/PatternSearch.hpp>


using namespace std;

namespace MapReduceTest
{

	const char* pattern = NULL;

	/**
	 * MapReduce user defined map function.
	 * @param	in_r is pointer to input_reader struct used to read mapper file.
	 * @param	im_w is pointer to intermediate_writer struct used to write mapping parts.
	 * @return	true if mapping completed without any errors.
	 */
	bool
	map(struct mapreduce_input_reader* const in_r, struct mapreduce_intermediate_writer* const im_w)
	{
		// Read input
		
		char inputBuffer[128];
		size_t inputBufferSize = (sizeof inputBuffer)/(sizeof inputBuffer[0]);

		const int readSuccess = mapreduce_read_input(in_r,&inputBuffer,&inputBufferSize); 

	
		if(not readSuccess)
			return false;
	
		// Write intermediate file

		char secondPart[64];
		char firstPart[64];

		size_t firstSize = 64;
		size_t secondSize = 64;

		copy(inputBuffer, inputBuffer+63, firstPart );
		copy(inputBuffer+64, inputBuffer+128, secondPart );

		int writeSuccess = mapreduce_write_part(im_w, &firstPart, &firstSize, 0);
	
		if(not writeSuccess)
			return false;
	
		writeSuccess = mapreduce_write_part(im_w,&secondPart, &secondSize, 1);

		if(not writeSuccess)
			return false;

  		return true;
	}

	/**
	 * MapReduce user defined reduce function.
	 * @param	im_r is pointer to intermediate_reader struct used to read mappend files.
	 * @param	out_w is pointer to output_writer struct used to write reduce result to output file.
	 * @return	true if mapping completed without any errors.
	 */
	bool
	reduce(struct mapreduce_intermediate_reader* const im_r, struct mapreduce_output_writer* const out_w)
	{

		char buffer[64];
		size_t bufferSize = (sizeof buffer)/(sizeof buffer[0]);

		const int readSuccess = mapreduce_read_intermediate(im_r,&buffer,&bufferSize); 

		if(not readSuccess)
			return false;

		int occurrence = 0;

		for(int i = 0; i < 64; ++i)
			if(buffer[i] == pattern[0])
				occurrence++;			

		
		for(int j = occurrence; j > 0; j--)
		{
			char seven = pattern[0];
			size_t size = sizeof '7';
			int write_success = mapreduce_write_out(out_w,&seven, &size);
	
			if(not write_success)
				return false;
		}

 		return true;
	}


	/**
 	 * Class default constructor, allocate memory.
	 */
	PatternSearch::PatternSearch(int argc,const char** argv)
	{
		inputs = argv +1;
		inputsNum = argc - 3;
		pattern = argv[argc-2];
		patternAmount = atoi(argv[argc-1]);
		outputs[0] = "out1.out";
		outputs[1] = "out2.out";

		prepareTest();
	}

	/**
	 * Class default constructor, free allocated memory.
	 */
	PatternSearch::~PatternSearch()
	{

	}

	/**
	 * Method is used to prepare MapReduce specificacd tion of test case.
	 */
	void
	PatternSearch::prepareTest()
	{
		const size_t outputsNum = (sizeof outputs)/(sizeof outputs[0]);

		spec.map = &map;
		spec.reduce = &reduce;
  		spec.inputs = inputs;
  		spec.outputs = outputs;
  		spec.inputs_num = inputsNum;
  		spec.outputs_num = outputsNum;
	}

	/**
	 * Method is used to run pattern search test. It is use MapReduce library to count appearance of
	 * defined pattern like some letter/word.
	 */
	void
	PatternSearch::runTest()
	{
		bool ok = mapreduce(&spec, &error);

		string a(pattern);
		string properResult = a;
		
		for(int i = 0; i < patternAmount-1; i++)
			properResult += a;
					
		string result;
		string result2;

		ifstream file;

		file.open("out1.out");
		file >> result;
		file.close();
		
		file.open("out2.out");
		file >> result2;
		file.close();

		result += result2;


		cout << "PatternSearchTest result:" << endl;
		cout << "Expected: " << properResult <<endl;
		cout << "Got: " << result << endl;
		if(properResult == result)
		{
			cout << "Results are expected" << endl;
		}
		else
		{
		  	cout << "Results are unexpected" << endl;
			ok = false;
		}
		cout << "PatternSearchTest " << (ok ? "ok" : "not ok") << endl;
	}
}

